from .logger import log
from .api import GlobalObject
from .utils.threading_utils import async_run

# 创建全局可以访问且可序列化的对象
g = GlobalObject()

__all__ = [
    "log",
    "g",
    "GlobalObject",
    "async_run"
]



import logging
import colorlog
import os


class Logger:
    # 日志级别：（critical > error > warning > info > debug）
    CRITICAL = logging.CRITICAL
    FATAL = logging.FATAL
    ERROR = logging.ERROR
    WARNING = logging.WARNING
    WARN = logging.WARN
    INFO = logging.INFO
    DEBUG = logging.DEBUG
    NOTSET = logging.NOTSET

    def __init__(self, name=None, file_name=None, console_level=None, file_level=None):
        """
        自定义日志对象
        :param name: logger名称，默认输出日志到控制台
        :param file_name: 输出文件路径，为None则不输出日志到文件
        :param console_level: 控制台日志级别
        :param file_level: 日志文件的日志级别
        """
        if name is None:
            self.name = 'default_pq_logger'
        else:
            self.name = name

        if console_level is None:
            console_level = self.DEBUG

        if file_level is None:
            file_level = self.ERROR  # # 默认只有error和critical级别才会写入日志文件

        # 获取logger对象
        self.logger = logging.getLogger(self.name)
        # 阻止日志消息从当前Logger向父级Logger传递
        self.logger.propagate = False

        # 指定最低日志级别：（critical > error > warning > info > debug）
        self.logger.setLevel(level=self.DEBUG)

        # 控制台输出不同级别日志颜色设置
        self.color_config = {
            'DEBUG': 'cyan',
            'INFO': 'green',
            'WARNING': 'yellow',
            'ERROR': 'red',
            'CRITICAL': 'purple',
        }

        # -------------------------
        # 输出到控制台
        # 日志格化字符串
        # -------------------------
        self.add_console_handler(level=console_level)

        # -------------------------
        # 输出到文件
        # -------------------------
        if file_name:
            self.add_file_handler(file_name=file_name, level=file_level)

    def set_name(self, name):
        """
        设置logger的名称，如果此名称的logger不存在，则创建一个新的logger
        如果新建的logger，需要通过：add_console_handler()或者add_file_handler()输出到控制台或者文件
        :param name: logger名称
        :return:
        """
        self.name = name
        # 获取logger对象
        self.logger = logging.getLogger(self.name)
        # 阻止日志消息从当前Logger向父级Logger传递
        self.logger.propagate = False

        # 指定最低日志级别：（critical > error > warning > info > debug）
        self.logger.setLevel(level=self.DEBUG)

    def set_level(self, level=logging.INFO):
        logging.getLogger(self.name).setLevel(level)

    def add_console_handler(self, level=logging.DEBUG):
        """
        输出到控制台
        :param level: 日志级别
        :return:
        """
        console_fmt = '%(log_color)s%(asctime)s: %(levelname)s %(message)s'
        console_formatter = colorlog.ColoredFormatter(fmt=console_fmt, log_colors=self.color_config)
        console_handler = logging.StreamHandler()
        console_handler.setFormatter(console_formatter)
        console_handler.setLevel(level)
        self.logger.addHandler(console_handler)

    def add_file_handler(self, file_name: str, level=logging.WARNING):
        """
        输出到文件
        :param file_name:
        :param level: 日志级别
        :return:
        """
        import json_log_formatter

        file_formatter = json_log_formatter.VerboseJSONFormatter()
        file_handler = logging.FileHandler(filename=file_name, mode='a', encoding='utf-8')
        file_handler.setFormatter(file_formatter)
        file_handler.setLevel(level)
        logging.getLogger(self.name).addHandler(file_handler)

    def remove_file_handler(self, file_name: str):
        """
        删除输出到文件的handler
        :param file_name:
        :return:
        """
        logger = logging.getLogger(self.name)
        # 遍历logger的handlers列表
        for handler in list(logger.handlers):
            # 检查handler是否是FileHandler且其文件名匹配给定的文件名
            if isinstance(handler, logging.FileHandler):
                # handler.baseFilename 为绝对路径
                if handler.baseFilename == os.path.abspath(file_name):
                    # 从logger的handlers列表中移除handler
                    logger.removeHandler(handler)
                    # 关闭handler以释放资源（可选）
                    handler.close()
                    # 如果确定不再需要handler，也可以删除对它的引用（可选）
                    del handler
                    break  # 如果只期望移除一个匹配的handler，则退出循环

    def isEnabledFor(self, level):
        """
        Is this logger enabled for the specified level?
        """
        return self.logger.isEnabledFor(level)

    def _log(self, level, msg, args, **kwargs):
        """
        Low-level logging routine which creates a LogRecord and then calls
        all the handlers of this logger to handle the record.
        """
        self.logger.log(level, msg, *args, **kwargs)

    def debug(self, msg, *args, **kwargs):
        """
        Log 'msg % args' with severity 'DEBUG'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.debug("Houston, we have a %s", "thorny problem", exc_info=1)
        """
        if self.isEnabledFor(self.DEBUG):
            self._log(self.DEBUG, msg, args, **kwargs)

    def info(self, msg, *args, **kwargs):
        """
        Log 'msg % args' with severity 'INFO'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.info("Houston, we have a %s", "interesting problem", exc_info=1)
        """
        if self.isEnabledFor(self.INFO):
            self._log(self.INFO, msg, args, **kwargs)

    def warning(self, msg, *args, **kwargs):
        """
        Log 'msg % args' with severity 'WARNING'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.warning("Houston, we have a %s", "bit of a problem", exc_info=1)
        """
        if self.isEnabledFor(self.WARNING):
            self._log(self.WARNING, msg, args, **kwargs)

    def error(self, msg, *args, **kwargs):
        """
        Log 'msg % args' with severity 'ERROR'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.error("Houston, we have a %s", "major problem", exc_info=1)
        """
        if self.isEnabledFor(self.ERROR):
            self._log(self.ERROR, msg, args, **kwargs)

    def exception(self, msg, *args, exc_info=True, **kwargs):
        """
        Convenience method for logging an ERROR with exception information.
        """
        self.error(msg, *args, exc_info=exc_info, **kwargs)

    def critical(self, msg, *args, **kwargs):
        """
        Log 'msg % args' with severity 'CRITICAL'.

        To pass exception information, use the keyword argument exc_info with
        a true value, e.g.

        logger.critical("Houston, we have a %s", "major disaster", exc_info=1)
        """
        if self.isEnabledFor(self.CRITICAL):
            self._log(self.CRITICAL, msg, args, **kwargs)


# 创建默认的全局logger
log = Logger(name='default_pq_logger')

# ---------------------------------------------------------
# 外部可以访问的列表
# ---------------------------------------------------------
__all__ = ["log", "Logger"]
__all__.extend([name for name in globals().keys() if name.startswith("get")])



# -*- coding: utf-8 -*-

import copy
from inspect import isclass
from pandas.compat.pickle_compat import pkl, Unpickler, load as _load


def load_reduce(self):
    stack = self.stack
    args = stack.pop()
    func = stack[-1]

    if len(args) and type(args[0]) is type:
        n = args[0].__name__  # noqa

    try:
        stack[-1] = func(*args)
        return
    except Exception as e:
        msg = "_reconstruct: First argument must be a sub-type of ndarray"

        if msg in str(e):
            try:
                cls = args[0]
                stack[-1] = object.__new__(cls)
                return
            except TypeError:
                pass

        if args and isclass(args[0]) and issubclass(args[0], BaseOffset):
            # TypeError: object.__new__(Day) is not safe, use Day.__new__()
            try:
                cls = args[0]
                stack[-1] = cls.__new__(*args)
                return
            except TypeError:
                pass

        # try to re-encode the arguments
        if getattr(self, "encoding", None) is not None:
            args = tuple(
                arg.encode(self.encoding) if isinstance(arg, str) else arg
                for arg in args
            )
            try:
                stack[-1] = func(*args)
                return
            except TypeError:
                pass

        # unknown exception, re-raise
        if getattr(self, "is_verbose", None):
            print(sys.exc_info())
            print(func, args)
        raise


class Unpickler2(Unpickler):
    pass


Unpickler2.dispatch = copy.copy(Unpickler2.dispatch)
Unpickler2.dispatch[pkl.REDUCE[0]] = load_reduce


def load(fh, encoding=None, is_verbose=True):
    try:
        return _load(fh, encoding=encoding)
    except (ValueError, TypeError) as ex:
        try:
            fh.seek(0)
            if encoding is not None:
                up = Unpickler2(fh, encoding=encoding)
            else:
                up = Unpickler2(fh)
            up.is_verbose = is_verbose

            return up.load()
        except Exception:
            raise ex



import datetime
from typing import Callable
from pqsdk.api import is_open_trade_date, get_next_trading_date
from pqsdk import log


class TimerFactory:
    def __init__(self, unit: str):
        """
        行情周期单位： 1d, 1m, 5m
        :param unit:
        """
        self.unit = unit      # 回测运行中基准行情的单位
        self.timer_list = []  # 定时器字典列表
        self.carry_dict = {}  # 顺延下一个交易日运行的timer {index: timer}
        self.timer_index = 0  # 给每个定时器一个编号

    @classmethod
    def format_time_str(cls, time_str):
        """
        字符串time_str作为输入，并返回格式化为HH:MM:SS的字符串，如果输入为HH:MM格式，则自动补充为HH:MM:00
        :param time_str:
        :return:
        """
        parts = time_str.split(':')
        if len(parts) == 2:
            # 如果是'HH:MM'格式，则自动补充为'HH:MM:00'
            return f"{parts[0]}:{parts[1]}:00"
            # 如果不是'HH:MM'格式，则检查是否是'HH:MM:SS'格式
        elif len(parts) == 3:
            # 如果是'HH:MM:SS'格式，则直接返回原字符串
            return time_str
        else:
            # 如果不是期望的格式，则返回错误或空字符串
            raise Exception(f"字符串格式不符合: {time_str}, 必须为HH:MM格式，或者HH:MM:SS的格式")

    @classmethod
    def is_valid_time(cls, start_time, t, interval: int, unit: str = 'm'):
        """
        判断给定时间t是否大于或等于开始时间start_time，并且t是以n个unit做为间隔的

        :param start_time: 检查的起始时间，datetime.datetime类型
        :param t: 被检查的时间，datetime.datetime类型
        :param interval: 单位间隔
        :param unit: 单位
        :return: bool
        """

        if unit not in ['m', 's']:
            raise Exception("unit is wrong, must be 'm' for minutes, or 's' for second.")
        # 确保t不小于start_time
        if t < start_time:
            return False

            # 计算t和start_time之间的时间差
        time_diff = t - start_time

        # 将时间差转换为分钟，并检查是否为n分钟的整数倍
        if unit == 'm':
            minutes_diff = time_diff.total_seconds() // 60  # 转换为分钟
        else:
            minutes_diff = time_diff.total_seconds()
        return minutes_diff % interval == 0

    def add_timer(self,
                  callback: Callable,
                  kwargs,
                  when='00:00:00',
                  weekdays=None,
                  weekcarry=False,
                  monthdays=None,
                  monthcarry=True,
                  dates=None,
                  minutes=None,
                  seconds=None,
                  datecarry=True):
        """
        添加计时器

        :param callback: 回调函数
        :param kwargs: 回调函数的参数
        :param when: 运行时间, format: '%H:%M:%S'
        :param weekdays: 每星期的第几天运行, Monday is 1, Sunday is 7
        :param weekcarry: weekday为非交易日，是否顺延下一个交易日
        :param monthdays: 每月的第几天运行
        :param monthcarry: 为非交易日，是否顺延下一个交易日
        :param dates: 间隔天数执行的日期列表，start_date为第一次执行日期
        :param minutes: 以when作为时间起点, 间隔的分神数执行
        :param seconds: 以when作为时间起点, 间隔的秒数执行
        :param datecarry: 为非交易日，是否顺延下一个交易日
        :return:
        """
        # 格式化为HH:MM:SS的字符串
        when = self.format_time_str(when)

        if monthdays is None:
            monthdays = []
        if weekdays is None:
            weekdays = []
        if dates is None:
            dates = []

        self.timer_index += 1
        timer = dict(index=self.timer_index,
                     callback=callback,
                     kwargs=kwargs,
                     when=when,
                     weekdays=weekdays,
                     weekcarry=weekcarry,
                     monthdays=monthdays,
                     monthcarry=monthcarry,
                     dates=dates,
                     minutes=minutes,
                     seconds=seconds,
                     datecarry=datecarry
                     )
        self.timer_list.append(timer)

    def carry_on(self, index, date, callback, kwargs):
        """
        顺延到下一个交易日执行的回调函数
        :param index: 定时器编号
        :param date:
        :param callback: 回调函数
        :param kwargs: 回调函数的参数字典
        :return:
        """
        self.carry_dict[index] = dict(date=date, callback=callback, kwargs=kwargs)

    def notify_timer(self, current_dt: datetime.datetime):
        """
        在回测中，此函数根据主图的Bar被触发，根据参数日期时间，遍历timer_list, 对符合条件的timer执行callback回调函数
        :param current_dt: 日期时间，datetime类型
        :return:
        """
        for timer in self.timer_list:
            index = timer['index']
            callback = timer['callback']
            kwargs = timer['kwargs']
            weekdays = timer['weekdays']
            weekcarry = timer['weekcarry']
            monthdays = timer['monthdays']
            monthcarry = timer['monthcarry']
            dates = timer['dates']
            datecarry = timer['datecarry']

            current_date = current_dt.strftime('%Y-%m-%d')

            # 1、先执行顺延的任务
            if index in self.carry_dict:
                carry_timer = self.carry_dict[index]
                # 检查是否顺延执行
                if current_date == carry_timer['date']:
                    # print(f"carry On: index={index}, date={carry_timer['date']}")
                    carry_timer['callback'](**carry_timer['kwargs'])
                    del self.carry_dict[index]  # 删除已经执行的顺延timer
                    # 碰巧当日也是顺延执行timer的执行日期, callback不需要重复执行
                    continue

            # 2、执行计划任务
            if not is_open_trade_date(current_date):  # 非交易日
                # 下一个交易日
                next_trade_date = get_next_trading_date(current_date)

                # 每周
                if len(weekdays) > 0 and current_dt.weekday() + 1 in weekdays and weekcarry:
                    log.warning(f"{current_dt} 非交易日, 顺延{callback.__name__}()函数执行日期: {next_trade_date} {timer['when']}")
                    self.carry_on(index, next_trade_date, callback, kwargs)  # 顺延下一个交易日执行

                # 每月
                if len(monthdays) > 0 and current_dt.day in monthdays and monthcarry:
                    log.warning(f"{current_dt} 非交易日, 顺延{callback.__name__}()函数执行日期: {next_trade_date} {timer['when']}")
                    self.carry_on(index, next_trade_date, callback, kwargs)  # 顺延下一个交易日执行

                # 每周期
                if len(dates) > 0 and current_date in dates and datecarry:
                    log.warning(f"{current_dt} 非交易日, 顺延{callback.__name__}()函数执行日期: {next_trade_date} {timer['when']}")
                    self.carry_on(index, next_trade_date, callback, kwargs)  # 顺延下一个交易日执行

                continue

            # 每周运行判断
            if len(weekdays) > 0 and current_dt.weekday() + 1 not in weekdays:
                continue

            # 每月运行判断
            if len(monthdays) > 0 and current_dt.day not in monthdays:
                continue

            # 每周期运行判断
            if len(dates) > 0 and current_date not in dates:
                continue

            # 执行回调函数的时机
            if self.unit in ['1d']:  # 日行情，每日执行
                # 如果是基准周期(self.unit)是1d则直接放行
                callback(**kwargs)
            elif self.unit in ['1m', '5m'] and timer['minutes'] is not None:  # 每间隔n分钟执行
                # 如果基准周期(self.unit)是1m或者5m, 间隔n分钟后执行
                start_time_str = f"{current_dt.strftime('%Y-%m-%d')} {timer['when']}"
                start_time = datetime.datetime.strptime(start_time_str, "%Y-%m-%d %H:%M:%S")
                if self.is_valid_time(start_time=start_time, t=current_dt, interval=timer['minutes'], unit='m'):
                    callback(**kwargs)
            elif self.unit in ['1m', '5m'] and timer['seconds'] is not None:  # 每间隔n秒执行
                # 如果基准周期(self.unit)是1m或者5m, 间隔n秒后执行

                # TODO 因为还没有秒级的行情数据，当秒级间隔时直接放行，否则使用下方注释掉的逻辑
                callback(**kwargs)

                # TODO 临时注释掉的逻辑，直到有秒级的数据
                # start_time_str = f"{current_dt.strftime('%Y-%m-%d')} {timer['when']}"
                # start_time = datetime.datetime.strptime(start_time_str, "%Y-%m-%d %H:%M:%S")
                # if self.is_valid_time(start_time=start_time, t=current_dt, interval=timer['seconds'], unit='s'):
                #     callback(**kwargs)

            elif self.unit in ['1m', '5m'] and current_dt.strftime('%H:%M:%S') == timer['when']:  # 分钟行情，指定时间执行
                # 如果基准周期(self.unit)是1m或者5m，则时间到达后执行
                callback(**kwargs)



from pqsdk.interface import AbstractRunInfo
from pqsdk.api import get_index_members_lst


class RunInfo(AbstractRunInfo):

    def __init__(self, kwargs: dict, context):
        self.__kwargs = kwargs
        self.context = context

    @property
    def tenant_id(self):
        return self.__kwargs.get('user').tenant_id

    @property
    def portfolio_id(self):
        """
        回测的时候，投资组合可能还未指定
        :return:
        """
        return -1

    @property
    def parameters(self):
        return self.__kwargs.get('parameters', {})

    @property
    def start_date(self):
        return self.__kwargs.get('start_date')

    @property
    def end_date(self):
        return self.__kwargs.get('end_date')

    @property
    def stock_starting_cash(self):
        return self.__kwargs.get('init_investment')

    @property
    def strategy_id(self):
        return self.__kwargs.get('strategy_id')

    @property
    def release_id(self):
        """
        回测的时候，策略代码可能还未发版
        :return:
        """
        return -1

    @property
    def strategy_name(self):
        return self.__kwargs.get('strategy_name')

    @property
    def strategy_remark(self):
        """
        策略备注
        :return:
        """
        return 'backtest'

    @property
    def is_queued_order(self):
        """
        回测中不拦截委托
        :return:
        """
        return False

    @property
    def benchmark(self):
        return self.__kwargs.get('benchmark')

    @property
    def stock_pool(self) -> list:
        return self.__kwargs.get('stock_pool')

    @property
    def stock_pool_members(self) -> list:
        """
        当日的股票池成员
        :return:
        """
        return get_index_members_lst(index_lst=self.stock_pool,
                                     trade_date=self.context.current_dt.strftime('%Y-%m-%d'))



# coding=utf-8
from functools import wraps
from ..utils import file_util as fu
import json
import os


def get_mac_address():
    import uuid
    mac = uuid.UUID(int=uuid.getnode()).hex[-12:].upper()
    return '%s:%s:%s:%s:%s:%s' % (mac[0:2], mac[2:4], mac[4:6], mac[6:8], mac[8:10], mac[10:])


def assert_auth(func):
    @wraps(func)
    def _wrapper(*args, **kwargs):
        from .client import PQDataClient
        # 如果未登录，登录远端并初始化
        if not PQDataClient.instance() or not PQDataClient.instance().inited:
            # client未初始化，尝试从环境变量或者配置文件登录并初始化
            config_file = 'config.sdk.json'
            if "AUTH_TOKEN" in os.environ:
                # 从环境变量的token登录
                PQDataClient.set_auth_params(token=os.environ.get("AUTH_TOKEN"),
                                             host=os.environ.get("AUTH_HOST"),
                                             port=os.environ.get("AUTH_PORT"),
                                             audience=os.environ.get("AUTH_AUDIENCE"))
            elif fu.check_path_exists(config_file):
                with open(config_file, 'r', encoding='utf-8') as f:
                    sdk_config = json.loads(f.read())
                # 从配置文件的token登录
                PQDataClient.set_auth_params(host=sdk_config['host'],
                                             token=sdk_config['token'],
                                             audience=sdk_config['audience'])

        # 检查是否登录成功
        if not PQDataClient.instance() or not PQDataClient.instance().inited:
            # 登录失败
            raise Exception("请先进行登录认证")
        else:
            return func(*args, **kwargs)

    return _wrapper



# coding=utf-8
from .main import BacktestExecutor
import pqsdk.utils.file_util as fu
import json
import pqsdk.api as pqdatasdk
from pqsdk import log
from pqsdk.api import PQDataClient
import pandas as pd
import quantstats as qs
from pathlib import Path
import datetime
import os


def execute(parameters: dict, script: str = None, strategy_file: str = None, task=None):
    """
    执行回测
    :param parameters: 回测参数字典
    :param script: 策略脚本
    :param strategy_file: 策略脚本文件
    :param task: celery task
    :return:
    """
    if not PQDataClient.instance() or not PQDataClient.instance().inited:
        log.debug("登录pqdatasdk服务端点")
        config_file = 'config.sdk.json'
        if not fu.check_path_exists(config_file):
            err_msg = "未找到配置文件. 请先运行命令进行配置: pqsdk config"
            log.error(err_msg)
            raise Exception(err_msg)

        with open(config_file, 'r', encoding='utf-8') as f:
            sdk_config = json.loads(f.read())
        pqdatasdk.auth_by_token(token=sdk_config['token'], host=sdk_config['host'], audience=sdk_config['audience'])

    if strategy_file:
        strategy_file = os.path.splitext(strategy_file)[0]
    kwargs = {"parameters": parameters, "script": script, "strategy_file": strategy_file}
    executor = BacktestExecutor(kwargs=kwargs, task=task)
    results = executor.run()
    return results


def tearsheet(results: dict, save_path="storage/reports"):
    """
    保存回测结果到Tearsheet

    :param save_path: 保存路径
    :param results:
    :return None: .
    """
    log.info("生成Tearsheet, 请稍后...")

    # 策略收益率
    strat_return = results['analysis']['time_return']['strat_return']
    strat_df = pd.DataFrame({'strat_return': strat_return})
    strat_df.index = pd.to_datetime(strat_df.index)
    strat_returns = strat_df['strat_return']

    # benchmark收益率
    bchmk_return = results['analysis']['time_return']['bchmk_return']
    bchmk_df = pd.DataFrame({'bchmk_return': bchmk_return})
    bchmk_df.index = pd.to_datetime(bchmk_df.index)
    # 重命名benchmark的Column名称，用于Tearsheet显示
    bchmk_df = bchmk_df.rename(columns={"bchmk_return": results['benchmark']})
    # bchmk_df = bchmk_df.loc[stats_df.index]  # 以持仓日期列表保留benchmark的记录
    bchmk_returns = bchmk_df[results['benchmark']]

    # tearsheet保存路径
    ts_path = Path(save_path)
    ts_path.mkdir(parents=True, exist_ok=True)
    time_str = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    filename = (
            "tearsheet"
            + "-"
            + time_str.replace("-", "").replace(":", "").replace(" ", "-")
            + ".html"
    )
    filepath = ts_path / filename

    title = f"回测报告 {time_str}"
    qs.reports.html(
        strat_returns,
        benchmark=bchmk_returns,
        title=title,
        output=filepath,
        download_filename=filepath
    )

    log.info(f"Tearsheet创建完成，路径：{filepath}")


def save_metrics(results: dict, save_path="storage/metrics"):
    """
    保存最终回测结果
    :param results:
    :param save_path: 保存路径
    :return:
    """
    # 计算总市值时间序列
    cost = results['analysis']['time_return']['cost']
    market_values = results['analysis']['time_return']['market_value']
    pnls = results['analysis']['time_return']['pnl']
    principal = results['analysis']['time_return']['principal']
    cash = results['analysis']['time_return']['cash']
    asset = results['analysis']['time_return']['asset']
    position_size = results['analysis']['time_return']['position_size']
    nav = results['analysis']['time_return']['nav']
    strat_return = results['analysis']['time_return']['strat_return']
    bchmk_returns = results['analysis']['time_return']['bchmk_return']
    metrics_df = pd.DataFrame({'cost': cost,  # 交易成本
                               'market_value': market_values,  # 总市值
                               'pnl': pnls,  # 盈亏
                               'principal': principal,  # 本金
                               'cash': cash,  # 现金
                               'asset': asset,  # 总资产
                               'position_size': position_size,  # 仓位
                               'nav': nav,  # 净值
                               'strat_return': strat_return,  # 策略收益率
                               'bchmk_return': bchmk_returns  # 基准收益率
                               })
    metrics_df.index = pd.to_datetime(metrics_df.index)
    metrics_df.index.name = "trade_date"
    metrics_df['cum_return'] = (1 + metrics_df['strat_return']).cumprod() - 1  # 累计收益率

    run_time = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    file_path = Path(save_path)
    file_path.mkdir(parents=True, exist_ok=True)
    filename = ("results" + "_" + run_time.replace("-", "").replace(":", "").replace(" ", "_") + ".csv")
    file_path = file_path / filename
    log.info(f"save metrics data: {file_path}")
    metrics_df.to_csv(file_path, index=True)


def save_orders(results: dict, save_path="storage/orders"):
    orders_df: pd.DataFrame = results['orders']
    run_time = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    file_path = Path(save_path)
    file_path.mkdir(parents=True, exist_ok=True)
    filename = ("orders" + "_" + run_time.replace("-", "").replace(":", "").replace(" ", "_") + ".csv")
    file_path = file_path / filename
    log.info(f"save orders data: {file_path}")
    orders_df.to_csv(file_path, index=False)


def save_inout_cash(results: dict, save_path="storage/inout_cash"):
    df: pd.DataFrame = results['inout_cash']
    run_time = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    file_path = Path(save_path)
    file_path.mkdir(parents=True, exist_ok=True)
    filename = ("inout_cash" + "_" + run_time.replace("-", "").replace(":", "").replace(" ", "_") + ".csv")
    file_path = file_path / filename
    log.info(f"save inout_cash data: {file_path}")
    df.to_csv(file_path, index=False)


# ---------------------------------------------------------
# 外部可以访问的列表
# ---------------------------------------------------------
__all__ = ["execute", "tearsheet", "save_metrics", "save_orders", "save_inout_cash"]



# coding=utf-8
from thriftpy2.transport import TTransportException

from .client import PQDataClient
from .utils import *
from typing import Union, List
import time
from ..logger import log
import pandas as pd
# 使用LRU缓存减少数据库查询
from functools import lru_cache


class GlobalObject(object):
    """
    全局变量对象
    如果g中的某个变量不想被序列化, 可以让变量以 '__' 开头, 在序列化时会被忽略

    要在Python中通过pickle序列化自定义类，需要满足以下两个条件：

    1. 自定义类必须实现__getstate__()和__setstate__()方法。
        1) getstate() 方法应该返回一个包含对象状态的字典。这个字典将是 pickle 用于序列化对象的内容。
        2) setstate() 方法接收一个字典作为参数，并使用它来恢复对象状态。
    2. 自定义类必须是全局可访问的。

    """

    def __getstate__(self):
        return {name: getattr(self, name) for name in dir(self) if not name.startswith("__")}

    def __setstate__(self, state):
        # reset g
        for k, v in state.items():
            setattr(self, k, v)


@assert_auth
def get_factor(sec_code: Union[None, str, List[str]] = None,
               factor: Union[None, str, List[str]] = None,
               sec_code_list: list = None,
               factor_list: list = None,
               factor_group_list: list = None,
               trade_date: str = None,
               start_date: str = None,
               end_date: str = None,
               start_datetime: str = None,
               end_datetime: str = None,
               stock_pool: list = None,
               dividend_type: str = 'front',
               unit: str = '1d',
               expect_df: bool = True,
               count: int = None,
               fill_data: bool = False):
    """
    查询因子库数据.


    :param fill_data: bool 是否向后填充空缺数据, 目前支持unit = '1d'
    :param sec_code: 股票代码，支持字符串或者字符串数组. 为数组的时候等价于sec_code_list
    :param factor: 因子名称, 支持字符串或者字符串数组，为数组的时候，等价于factor_list
    :param sec_code_list: 股票代码列表
    :param factor_list: 因子列表，因子可能来源于不同的因子分组
    :param factor_group_list: 限定因子分组
    :param trade_date: 限定交易日
    :param start_date: 开始交易日
    :param end_date: 结束交易日
    :param start_datetime: 当unit='1m'时有效，格式：2023-03-29 14:57:00
    :param end_datetime: 当unit='1m'时有效，格式：2023-03-29 14:57:00
    :param stock_pool: 指数列表，仅加载指数成分股。
                       选出了历史中出现在指数中所有成分，此时有未来函数，等于把低价股票提前选入了沪深300的股票池中。
                       数据分析中需要结合crud.get_index_members_lst()一起使用，过滤掉未来函数的股票。
    :param dividend_type: 复权选项(对股票/基金的价格字段、成交量字段及factor字段生效)
                'front'
                : 前复权, 默认是前复权
                none
                : 不复权, 返回实际价格
                'back'
                : 后复权
    :param unit: 单位时间长度，支持1d、1m，默认为1d
    :param expect_df: 是否以DataFrame返回
    :param count: 数量, 返回最近的条目数
    :return:
    """
    if factor_list is None:
        factor_list = ['close']
    return PQDataClient.instance().get_factor(**locals())


@assert_auth
def get_transform_factor(sec_code: str = None,
                         sec_code_list: list = None,
                         factor_list=None,
                         factor_group_list: list = None,
                         trade_date: str = None,
                         start_date: str = None,
                         end_date: str = None,
                         stock_pool: list = None,
                         dividend_type='back',
                         unit: str = '1d',
                         expect_df: bool = True,
                         count: int = None):
    """
    查询因子数据的空值处理、去极值、标准化、市值中性化、行业中性化后的结果

    :param sec_code: 股票代码
    :param sec_code_list: 股票代码列表
    :param factor_list: 因子列表，因子可能来源于不同的因子分组
    :param factor_group_list: 限定因子分组
    :param trade_date: 限定交易日
    :param start_date: 开始交易日
    :param end_date: 结束交易日
    :param stock_pool: 指数列表，仅加载指数成分股。
                       选出了历史中出现在指数中所有成分，此时有未来函数，等于把低价股票提前选入了沪深300的股票池中。
                       数据分析中需要结合crud.get_index_members_lst()一起使用，过滤掉未来函数的股票。
    :param dividend_type: 复权选项
                none
                : 不复权, 返回实际价格
                'back'
                : 后复权
    :param unit: 单位时间长度，支持1d、1m，默认为1d
    :param expect_df: 是否以DataFrame返回
    :param count: 数量, 返回的结果集的行数
    :return:
    """
    if factor_list is None:
        factor_list = ['close']
    return PQDataClient.instance().get_transform_factor(**locals())


@assert_auth
def get_history(count: int = None,
                unit: str = '1d',
                start_date: str = None,
                end_date: str = None,
                start_datetime: str = None,
                end_datetime: str = None,
                field: str = 'close',
                security_list=None,
                stock_pool: list = None,
                expect_df=True,
                skip_paused=False,
                dividend_type='front'):
    """
    获取历史数据，可查询多个标的单个数据字段，返回数据格式为 DataFrame 或 Dict(字典)

    :param stock_pool:
    :param start_date:
    :param end_date:
    :param count: 数量, 返回的结果集的行数
    :param unit: 单位时间长度, 几天或者几分钟, 现在支持'1d','1m'
    :param start_datetime: 当unit='1m'时有效，格式：2023-03-29 14:57:00
    :param end_datetime: 当unit='1m'时有效，格式：2023-03-29 14:57:00
    :param field: 要获取的数据字段
    :param security_list: 要获取数据的股票列表
    :param expect_df: expect_df=True: [pandas.DataFrame]对象, 行索引是[datetime.datetime]对象, 列索引是股票代号
               expect_df=False: dict, key是股票代码, value是一个numpy数组[numpy.ndarray]
    :param skip_paused: 是否跳过不交易日期(包括停牌, 未上市或者退市后的日期)
    :param dividend_type: 复权选项(对股票/基金的价格字段、成交量字段及factor字段生效)
                'front'
                : 前复权, 默认是前复权
                none
                : 不复权, 返回实际价格
                'back'
                : 后复权
    :return:
    """
    return PQDataClient.instance().get_history(**locals())


@assert_auth
def get_attribute_history(security: str,
                          count: int = None,
                          unit: str = '1d',
                          start_date: str = None,
                          end_date: str = None,
                          start_datetime: str = None,
                          end_datetime: str = None,
                          fields: list = None,
                          skip_paused: bool = True,
                          expect_df: bool = True,
                          dividend_type: str = 'front'):
    """
    获取历史数据，可查询单个标的多个数据字段，返回数据格式为 DataFrame 或 Dict(字典)

    :param start_date:
    :param end_date:
    :param start_datetime: 当unit='1m'时有效，格式：2023-03-29 14:57:00
    :param end_datetime: 当unit='1m'时有效，格式：2023-03-29 14:57:00
    :param security: 股票代码
    :param count: 数量, 返回的结果集的行数
    :param unit: 单位时间长度, 1d, 1m
    :param fields: 股票属性的list, 支持：['open', ' high', 'low', 'close', 'volume', 'amount']
    :param skip_paused: 是否跳过不交易日期(包括停牌, 未上市或者退市后的日期).
    :param expect_df: 若是True, 返回[pandas.DataFrame], 否则返回一个dict, 具体请看下面的返回值介绍. 默认是True.
    :param dividend_type: 复权选项(对股票/基金的价格字段、成交量字段及factor字段生效)
                'front'
                : 前复权, 默认是前复权
                none
                : 不复权, 返回实际价格
                'back'
                : 后复权
    :return:
    """
    return PQDataClient.instance().get_attribute_history(**locals())


@assert_auth
def get_synthesize_factor(stock_pool: list,
                          start_date: str,
                          end_date: str,
                          freq: int = 5,
                          trade_date: str = None,
                          sec_code: str = None,
                          sec_code_list: list = None,
                          factor_list=None,
                          expect_df: bool = True,
                          exact_match: bool = False):
    """
    获取参数组合（stock_pool, start_date, end_date, freq）最优的合成因子数据

    :param stock_pool: 股票池
    :param start_date: 因子取数的开始日期，也是合成因子回测的开始日期
    :param end_date: 因子取数的结束日期，也是合成因子回测的结束日期
    :param freq: 调仓周期，单位为天，默认值为5
    :param trade_date: 限定交易日
    :param sec_code: 股票
    :param sec_code_list: 股票代码列表
    :param factor_list: 合成因子列表，默认返回完整的合成因子列表
    :param expect_df: 若是True, 返回[pandas.DataFrame], 否则返回一个dict, 具体请看下面的返回值介绍. 默认是True.
    :param exact_match: 是否精确匹配（stock_pool, start_date, end_date, freq），或者近似匹配
    :return:
    """
    return PQDataClient.instance().get_synthesize_factor(**locals())


@assert_auth
def get_synthesize_portfolio(count: int = None,
                             stock_pool: list = None,
                             start_date: str = None,
                             end_date: str = None,
                             freq: int = 5,
                             trade_date_lst: list = None,
                             sec_code: str = None,
                             sec_code_list: list = None,
                             factor_list=None,
                             expect_df: bool = False,
                             exact_match: bool = False,
                             stock_size: int = 10):
    """
    获取参数组合（stock_pool, start_date, end_date, freq）最优的合成因子数据所产生的股票组合。

    :param count: 获取最新的n期调仓日数据
    :param stock_pool: 股票池
    :param start_date: 因子取数的开始日期，也是合成因子回测的开始日期
    :param end_date: 因子取数的结束日期，也是合成因子回测的结束日期
    :param freq: 调仓周期，单位为天，默认值为5
    :param trade_date_lst: 限定调仓日列表
    :param sec_code: 股票
    :param sec_code_list: 股票代码列表
    :param factor_list: 合成因子列表，默认返回完整的合成因子列表
    :param expect_df: 若是True, 返回[pandas.DataFrame], 否则返回一个dict, 具体请看下面的返回值介绍. 默认是True.
    :param exact_match: 是否精确匹配（stock_pool, start_date, end_date, freq），或者近似匹配
    :param stock_size: 股票组合的股票个数
    :return:
    """
    return PQDataClient.instance().get_synthesize_portfolio(**locals())


@assert_auth
@lru_cache(maxsize=1024)
def get_trade_cal(start_date: str = '2018-01-01'):
    """
    获取交易日历，默认从2018-01-01开始
    :return:
    """
    return PQDataClient.instance().get_trade_cal(**locals())


@assert_auth
@lru_cache(maxsize=1024)
def get_previous_trading_date(date: str, lookback_days: int = 1):
    """
    对于给定的日期，返回前第n天
    :param date:
    :param lookback_days: 前第n天，默认前1天
    :return:
    """
    return PQDataClient.instance().get_previous_trading_date(**locals())


@assert_auth
@lru_cache(maxsize=1024)
def get_next_trading_date(date: str, lookahead_days: int = 1):
    """
    对于给定的日期，返回后第n天
    :param date:
    :param lookahead_days: 后第n天，默认前1天
    :return:
    """
    return PQDataClient.instance().get_next_trading_date(**locals())


@assert_auth
@lru_cache(maxsize=1024)
def get_near_trade_date(trade_date: str, max_retries=5, retry_delay=2):
    """
    检查trade_date是否为交易日，否则，获取上一个交易日。
    如果返回值不是字符串类型，则重试，最多重试max_retries次，每次间隔retry_delay秒。

    :param trade_date: 输入的交易日期
    :param max_retries: 最大重试次数
    :param retry_delay: 每次重试之间的延迟时间（秒）
    :return: 最近的交易日（字符串类型）
    """
    retries = 0
    while retries < max_retries:
        res = PQDataClient.instance().get_near_trade_date(trade_date=trade_date)
        if isinstance(res, str):
            return res
        else:
            log.warning(f"返回值不是字符串类型，{res}，重试 {retries + 1}/{max_retries}...")
            retries += 1
            time.sleep(retry_delay)

            # 如果达到最大重试次数仍然失败，可以抛出一个异常或返回一个默认值
    raise ValueError("无法获取有效的交易日期，已达到最大重试次数")


@assert_auth
@lru_cache(maxsize=1024)
def get_open_trade_dates(start_date: str, end_date: str = None):
    """
    通过开始日期和结束日期，取得交易日。

    :param start_date:
    :param end_date:
    :return: df, 只有一个Column：trade_date
    """
    return PQDataClient.instance().get_open_trade_dates(**locals())


@assert_auth
def get_index_members_lst(index_lst: list, trade_date: str = None):
    """
    获取指数成分股列表

    :param trade_date: 截止trade_date最新一期的指数成分股，为None则返回历史出现过的所有成分股
    :param index_lst: 指数列表
    :return:
    """
    return PQDataClient.instance().get_index_members_lst(**locals())


@assert_auth
def get_concept_members(code: str = None, code_list: list = None, name: str = None, keyword: str = None):
    """
    获取概念成分股

    :param code: 概念编码
    :param code_list: 概念编码列表
    :param name: 概念名称
    :param keyword: 概念名称关键字
    :return: list 成分股列表
    """
    return PQDataClient.instance().get_concept_members(**locals())


@assert_auth
def get_concept(code: str = None,
                code_list: list = None,
                name: str = None,
                keyword: str = None,
                exchange: str = 'A',
                type: str = 'N') -> dict:
    """
    获取概念列表
    :param code: 概念编码
    :param code_list: 概念编码列表
    :param name: 概念名称
    :param keyword: 概念名称关键字
    :param exchange: 交易所
    :param type: 	N概念指数, S特色指数
    :return: dict {'864006.TI': {'name': '固态电池', 'count': 1}, '886032.TI': {'name': '固态电池', 'count': 62}}
    """
    return PQDataClient.instance().get_concept(**locals())


@assert_auth
def get_hot_concept(count: int,
                    trade_date: str,
                    code: str = None,
                    code_list: list = None,
                    name: str = None,
                    keyword: str = None,
                    exchange: str = 'A',
                    type: str = 'N',
                    order_by: str = 'pct_change desc'):
    """
    获取热门概念编号列表，默认以涨跌幅倒序排列
    :param count:
    :param trade_date:
    :param code:
    :param code_list:
    :param name:
    :param keyword:
    :param exchange:
    :param type:
    :param order_by:
    :return:
    """
    return PQDataClient.instance().get_hot_concept(**locals())


@assert_auth
def get_news_content(start_date: str, end_date: str, sec_code_list: list = None):
    return PQDataClient.instance().get_news_content(**locals())


@assert_auth
def get_stock_info(stock_lst: list,
                   fields=None,
                   stock_pool: list = None,
                   trade_date: str = None,
                   dividend_type: str = "front"):
    """
    获取股票基础信息。

    :param dividend_type: 对于历史数据，复权选项(对股票/基金的价格字段、成交量字段及factor字段生效)
                'front'
                : 前复权, 默认是前复权
                none
                : 不复权, 返回实际价格
                'back'
                : 后复权
    :param trade_date: 指定交易日
    :param stock_pool: 指定股票池
    :param stock_lst: 指定股票列表
    :param fields: 可选，['sec_code', 'datetime', 'name', 'pre_close', 'up_limit', 'down_limit']
    :return: df:
        se_code	股票代码
        datetime 数据时间
        name 	合约名称
        pre_close 	前收盘价格
        up_limit	当日涨停价
        down_limit	当日跌停价
    """
    return PQDataClient.instance().get_stock_info(**locals())


@assert_auth
def get_suspend_info(trade_date: str, fields=None, stock_lst: list = None, suspend_type: str = 'S'):
    """
    获取股票在某日的停复牌状态

    :param trade_date:
    :param fields:
    :param stock_lst:
    :param suspend_type: 停牌：S，复牌：R
    :return:
    """
    return PQDataClient.instance().get_suspend_info(**locals())


@assert_auth
def get_stock_name_his(sec_code: str, trade_date: str):
    """
    获取股票代码的历史名称
    :param trade_date:
    :param sec_code:
    :return: {'sec_code': '000001.SZ', 'name': '平安银行', 'start_date': '2012-08-02', 'end_date': '2099-12-30'}
    """
    return PQDataClient.instance().get_stock_name_his(**locals())


@assert_auth
@lru_cache(maxsize=1024)
def is_open_trade_date(date: str):
    """检查是否为交易日"""
    return PQDataClient.instance().is_open_trade_date(**locals())


@assert_auth
def get_last_ticks(sec_code_lst: list, index: bool = False, fields: list = None):
    """
    获取当前最新的Tick数据。
    注意：不是历史某个回测时间点的Tick数据。

    :param sec_code_lst:
    :param fields: 返回的字段列表
    :param index: 是否设置sec_code为Index
    :return:
            [{
            'sec_code', 股票代码
            ’datetime'  tick时间字符串： %Y-%m-%d %H:%M:%S
            'open', 当日开盘价
            'high', 当日收盘价
            'low', 当日最低价
            'last_price', 当日最新价格
            'pre_close', 昨日收盘价
            'volume', 当日成交量，单位：手
            'amount', 当日成交额
            }]
    """
    return PQDataClient.instance().get_last_ticks(**locals())


@assert_auth
def get_plate_data(trade_date: str = None, end_time: str = "15:00", limit: int = 3):
    """
    获取板块列表

    :param trade_date: 指定交易日的板块列表
    :param end_time: 指定板块排名的时间节点，从09:25开始到15:30结束，每个间隔5分钟
    :param limit: 按照板块强度倒序排名的板块数量
    :return:
    """
    return PQDataClient.instance().get_plate_data(**locals())


@assert_auth
def get_hot_plate(trade_date: str = None, n_days: int = 5, top_n: int = 5, hot_n: int = 5, exclude_list: list = None):
    """
    获取最近n天 上榜(每天进入强度前n)天数最多的板块
    :param trade_date: 选股日期, 默认为前一个交易日
    :param n_days: 从最近多少天里筛选热门板块，默认 5
    :param top_n:  每天热度top n，默认top 5
    :param hot_n:  最终选出 最近n_days 天 热门的前n个板块
    :param exclude_list: 排除的板块清单，比如 ST板块，中报增长板块等
    :return: DataFrame
    """
    return PQDataClient.instance().get_hot_plate(**locals())


@assert_auth
def get_plate_members_lst(plate_code_list: list, trade_date: str = None):
    """
    获取板块的成分股
    :param plate_code_list:
    :param trade_date:
    :return:
    """
    return PQDataClient.instance().get_plate_members_lst(**locals())


@assert_auth
def get_member_plates(sec_code: str, trade_date: str = None) -> pd.DataFrame:
    """
    获取股票代码所属的板块, 一个股票可能属于多个板块。除了获取板块外，额外提供板块强度，并且按照板块强度倒序排列。
    :param sec_code: 股票代码
    :param trade_date: 交易日，不同交易下的板块可能不同
    :return: DataFrame， 字段包括 板块代码，板块名称，板块强度
           plate_code plate_name  intensity
    0      801660         通信       2306
    1      801250       并购重组       1834
    2      801218       华为概念       1610
    3      801328       消费电子        640
    4      801001         芯片        456
    5      801878       端侧AI        296
    6      801408       智能家居        246
    7      801519        汽车类        227
    8      801857      PPP概念        136
    9      801033       国有企业         44
    10     801382     分拆上市预期         38
    """
    return PQDataClient.instance().get_member_plates(**locals())


@assert_auth
def get_stock_name(sec_code: str, trade_date: str = None):
    """
    获取股票代码的最新名称

    :param sec_code:
    :param trade_date:
        None：默认为当天，如果当天不是交易日，则获取上一个交易日。
        'YYYY-mm-dd'：具体日期的最新股票名称
    :return: 股票名称
    """
    return PQDataClient.instance().get_stock_name(**locals())


@assert_auth
def get_stock_price(sec_code: str,
                    trade_date: str = None,
                    end_time: str = None,
                    dividend_type: str = "front"):
    """
    获取最新的股票价格

    :param dividend_type: 复权选项(对股票/基金的价格字段、成交量字段及factor字段生效)
                'front'
                : 前复权, 默认是前复权
                none
                : 不复权, 返回实际价格
                'back'
                : 后复权
    :param trade_date: 指定交易日, 格式: 2024-07-15
    :param end_time: 如果trade_date不是系统日期当日，可以指定历史数据的具体分钟，格式：09:30
    :param sec_code:
    :return:
    """
    return PQDataClient.instance().get_stock_price(**locals())


@assert_auth
def get_price_limit(sec_code: str,
                    trade_date: str = None,
                    dividend_type: str = "front"):
    """
    股票代码涨跌停价格
    假设：
    - 600xxx, 601xxx, 603xxx：上海主板，涨跌停限制10%
    - 000xxx：深圳主板，涨跌停限制10%
    - 002xxx：深圳中小板，涨跌停限制10%
    - 300xxx：深圳创业板，涨跌停限制20%
    - 688xxx：上海科创板，涨跌停限制20%

    :param dividend_type: 复权选项(对股票/基金的价格字段、成交量字段及factor字段生效)
                'front'
                : 前复权, 默认是前复权
                none
                : 不复权, 返回实际价格
                'back'
                : 后复权
    :param sec_code:
    :param trade_date: 指定日期的涨跌停价格
    :return: (昨日收盘价格, 今日涨停价, 今日跌停价)
    """
    return PQDataClient.instance().get_price_limit(**locals())


@assert_auth
def get_auction_data(security: str,
                     count: int = None,
                     unit: str = '1m',
                     trade_date: str = None,
                     start_datetime: str = None,
                     end_datetime: str = None,
                     fields=None):
    """
    获取股票的集合竞价数据。

    :param security: 股票代码
    :param count: 返回行数
    :param unit: 支持1m, 5m
    :param trade_date: 交易日
    :param start_datetime: 开始时间
    :param end_datetime: 结束时间
    :param fields: 获取字段，范围：['open', 'high', 'low', 'close', 'pre_close']
    :return:
    """
    return PQDataClient.instance().get_auction_data(**locals())


@assert_auth
def get_abnormal_stocks(fields: list = None,
                        sec_code_list: list = None,
                        start_date: str = None,
                        end_date: str = None,
                        abnormal_type: int = None):
    """
    获取严重异动的股票，即其偏离值异动的股票
    :param fields: 获取数据的字段，默认：sec_code, 支持 sec_code, trade_date, abnormal_type
    :param sec_code_list: 指定的股票是否发生过异动
    :param start_date: 异动日期范围, 如果未指定起始日期，默认6个月前
    :param end_date: 异动日期范围
    :param abnormal_type: 股票偏离值异动类型
    :return:
    """
    return PQDataClient.instance().get_abnormal_stocks(**locals())


@assert_auth
def get_suspend_status(sec_code: str, trade_date: str = None):
    """
    合约停牌状态
    :param sec_code:
    :param trade_date:
    :return: int
            <=0 正常，-1 复牌，>0 停牌, >=1 停牌天数
    """
    return PQDataClient.instance().get_suspend_status(**locals())


@assert_auth
def get_prompt_query(query: str,
                     query_type: str = 'stock',
                     page_num: int = None,
                     page_size: int = None,
                     loop: Union[bool, int] = False):
    """
    通过Prompt Query的方式进行选股。
    举例：2024-10-16,DDE大单净量>0.25, 非科创板,非创业板,非北交所,非ST
    将获取相关的股票列表：
        ['002793.SZ', '000838.SZ', '600837.SH', '000972.SZ', '600889.SH', ...]

    :param query: 通过文本的方式进行选股
    :param query_type: 支持: stock: 股票，fund: 基金
    :param page_num: 可选，查询的页号，默认为1
    :param page_size: 可选，每页行数, 最小50， 最大100，默认为100
    :param loop: 是否循环分页，返回多页合并数据。默认值为False，可以设置为True或具体数值. 当为数值的时候，1-n的多页数据一起返回
    :return: list, 股票代码列表
    """
    params_dict = {"query": query,
                   "query_type": query_type,
                   "page_size": page_size,
                   }

    if page_num and loop is False:
        params_dict['page_num'] = page_num
        max_retry = 2  # 最多尝试次数
        result_on_except = []  # 如果尝试失败的返回默认结果
        return PQDataClient.instance().get_prompt_query(*(max_retry, result_on_except), **params_dict)

    elif isinstance(loop, bool) and loop is True:  # 分页查询，最终合并结果
        _page_num = 1
        result = []
        while True:
            params_dict['page_num'] = _page_num
            max_retry = 2  # 最多尝试次数
            result_on_except = []  # 如果尝试失败的返回默认结果
            stock_list = PQDataClient.instance().get_prompt_query(*(max_retry, result_on_except), **params_dict)
            result.extend(stock_list)
            _page_num += 1
            if len(stock_list) == 0:
                break
        return result
    elif isinstance(loop, int):  # 返回前x页
        loop = max(1, loop)
        params_dict['loop'] = loop
        return PQDataClient.instance().get_prompt_query(**params_dict)
    else:
        return PQDataClient.instance().get_prompt_query(**params_dict)


@assert_auth
def get_reduce_holdings(start_date: str, end_date: str):
    """
    获取指定日期范围内发布了减持公告的股票代码列表

    :param start_date: 开始日期
    :param end_date: 结束日期
    :return: list，股票代码列表
    """
    return PQDataClient.instance().get_reduce_holdings(**locals())


@assert_auth
def get_board_members(boards: list,
                      status: str = "L",
                      start_date: str = None,
                      end_date: str = None,
                      is_hs: str = None):
    """
    获取板块的成分股

    :param boards: 板块列表
            [
            "sz.zh", # 深市主板
            "sh.zh", # 沪市主板
            "sz.cy", # 创业板
            "sh.kc", # 科创板
            "sz.zx",  # 中小板
            "bj.bj",  # 北交所
            ]
    :param status: 上市状态 L上市 D退市 P暂停上市
    :param start_date: # 上市日期范围
    :param end_date: # 上市日期范围
    :param is_hs: 是否沪深港通标的，N否 H沪股通 S深股通
    :return: list, 股票列表
    """
    return PQDataClient.instance().get_board_members(**locals())


@assert_auth
@lru_cache(maxsize=1024)
def get_up_limit_stocks(stock_list: list, trade_date: str, lookback_days: int = 1, dividend_type: str = 'front'):
    """
    获取连板涨停的股票列表。

    :param dividend_type: 复权选项(对股票/基金的价格字段、成交量字段及factor字段生效)
                'front'
                : 前复权, 默认是前复权
                none
                : 不复权, 返回实际价格
                'back'
                : 后复权
    :param stock_list:
    :param trade_date:
    :param lookback_days:
            0：昨日未涨停的股票
            >=1: 连板涨停的股票
    :return: list，股票代码列表
    """
    return PQDataClient.instance().get_up_limit_stocks(**locals())


# ---------------------------------------------------------
# 外部可以访问的列表
# ---------------------------------------------------------
__all__ = ["GlobalObject", "is_open_trade_date"]
__all__.extend([name for name in globals().keys() if name.startswith("get")])



# coding=utf-8
import argparse
import sys
from .main import run_backtest
from pqsdk.utils.value_type_util import convert_to_type
import requests
import getpass
import pqsdk.utils.file_util as fu
import json
from .logger import log
import logging

log.set_level(logging.INFO)


def input_with_default(prompt, default):
    user_input = input(prompt)
    return user_input if user_input else default


def handle_config_command():
    # 提示用户输入用户名
    username = input("请输入用户名: ")
    if not username:
        print("用户名不能为空！")
        return

        # 提示用户输入密码
    password = getpass.getpass("请输入密码: ")
    if not password:
        print("密码不能为空！")
        return

    host = 'api.pinkquant.com'

    # 获取Tenant list
    url = f"http://{host}/api/admin/tenant/login/tenants?username={username}"
    response = requests.get(url)
    tenants = response.json()['data']

    if len(tenants) == 0:
        print(f"用户 {username} 未找到有效的协作空间, 请联系管理员")
        return

    if len(tenants) == 1:
        # 仅归属一个协作空间
        tenant_index = 0
    else:
        print(f"用户 {username} 归属于如下的协作空间: \n")
        for index, tenant in enumerate(tenants):
            print(f"{index}: ", tenant['enterpriseName'])

        tenant_index = input_with_default("\n请输入协作空间的编号(默认值为0): ", "0")

    selectedTenant = tenants[int(tenant_index)]
    tenant_id = selectedTenant['tenantCode']

    # 获取token
    url = f'http://{host}/api/auth/oauth2/token'
    params = {
        'grant_type': 'password',
        'tenantId': tenant_id
    }
    headers = {
        'Authorization': 'Basic YXBpLWNsaWVudDoxMjM0NTY=',
        'Content-Type': 'application/x-www-form-urlencoded'
    }
    response = requests.post(url, params=params, headers=headers, data={"username": username, "password": password})
    result = response.json()
    if 'data' not in result:
        log.error(f"用户未分配正确权限，请联系管理员")
        return

    result = result['data']

    config_file = 'config.sdk.json'
    if fu.check_path_exists(config_file):
        # 更新配置文件
        with open(config_file, 'r', encoding='utf-8') as f:
            sdk_config = json.loads(f.read())
        sdk_config['token'] = result['access_token']
        sdk_config['audience'] = "api-client"
        sdk_config['host'] = host
        with open(config_file, 'w', encoding='utf-8') as f:
            json.dump(sdk_config, f)
        print("配件文件已经更新")
    else:
        # 创建配置文件
        data = {
            "host": host,
            "audience": "api-client",
            "token": result['access_token']
        }
        with open(config_file, 'w', encoding='utf-8') as f:
            json.dump(data, f)
        print("新的配置创建完成")

    # 检查响应状态码
    if response.status_code == 200:
        pass
    else:
        print(f"请求失败，状态码：{response.status_code}")


def handle_config_info_command():
    config_file = 'config.sdk.json'
    if fu.check_path_exists(config_file):
        with open(config_file, 'r', encoding='utf-8') as f:
            sdk_config = json.loads(f.read())

        print("配置信息:")
        for k, v in sdk_config.items():
            print(f"    {k} = {v}")
    else:
        print("未找到配置文件. 请先运行命令进行配置: pqsdk config")


def main():
    parser = argparse.ArgumentParser(description='PQ SDK Command Line Interface')
    parser.add_argument('command', choices=['backtest', 'config'], help='The command to execute')
    parser.add_argument('subcommand', choices=['info'], nargs='?', help='The subcommand to execute (optional)')
    parser.add_argument('-f', dest='file', metavar='file path', help='回测脚本文件路径')
    parser.add_argument('--plot', action='store_true', help='是否绘图')
    parser.add_argument('--save_metrics', action='store_true', help='是否保存结果到csv文件')
    parser.add_argument('--save_orders', action='store_true', help='是否保存委托到csv文件')
    parser.add_argument('--save_inout_cash', action='store_true', help='是否保存出入金到csv文件')
    parser.add_argument('-p', dest='param', action='append', nargs=1, metavar='key=value',
                        help='自定义的参数, 格式: -p key=value, 等号之间不能有空格')

    args = parser.parse_args()

    # 将 -p 传递的参数转换为字典, args.param的原始格式：[['key1=value1'], ['key2=value2']]
    params = {}
    if args.param is not None:
        for kv in args.param:
            key, value = kv[0].replace(" ", "").split('=')
            params[key] = convert_to_type(value)

    if args.command == 'backtest':
        # 执行回测
        run_backtest(args.file, params, args.plot, args.save_metrics, args.save_orders, args.save_inout_cash)
    elif args.command == 'config':
        if args.subcommand == 'info':
            handle_config_info_command()
        else:
            handle_config_command()

    else:
        parser.print_help()
        sys.exit(1)


if __name__ == "__main__":
    main()



import itertools
import random
import threading
from thriftpy2 import transport, protocol
from thriftpy2.rpc import make_client
from .thrift_client import thrift
from .utils import get_mac_address
from ..version import __version__ as current_version
import platform
import time
import socket
import six
import msgpack
import zlib
from .exceptions import ResponseError
import pandas as pd
from .compat import pickle_compat as pc
import json
from ..logger import log
from .request import RequestHandler

if platform.system().lower() != "windows":
    socket_error = (transport.TTransportException, socket.error, protocol.cybin.ProtocolError)
else:
    socket_error = (transport.TTransportException, socket.error)


class PQDataClient(object):
    _lock = threading.Lock()  # 类级别的锁，用于线程安全
    _instances = {}  # 类级别的字典，用于存储单例实例
    _auth_params = {}

    _default_host = "api.pinkquant.com"
    _default_port = 7500

    request_timeout = 300
    request_attempt_count = 3

    def __init__(self,
                 host=None,
                 port=None,
                 username="",
                 password="",
                 token="",
                 token_type="",
                 audience="resource-client"):
        self.host = host or self._default_host
        self.port = int(port or self._default_port)
        self.username = username
        self.password = password
        self.token = token
        self.audience = audience
        self.token_type = token_type

        assert self.host, "host is required"
        assert self.port, "port is required"
        assert self.username or self.token, "username is required"
        assert self.password or self.token, "password is required"

        self.client = None
        self.inited = False
        self.not_auth = True
        self.compress = True
        self.data_api_url = ""
        self._http_token = ""
        self.request_handler = RequestHandler()

        self._request_id_generator = itertools.count(random.choice(range(0, 1000, 10)))

    @classmethod
    def set_auth_params(cls, **params):
        if params != cls._auth_params and cls.instance():
            cls.instance()._reset()
            cls._instances['_instance'] = None
        cls._auth_params = params
        cls.instance().ensure_auth()

    def _reset(self):
        if self.client:
            self.client.close()
            self.client = None
        self.inited = False
        self.http_token = ""

    def _create_client(self):
        self.client = make_client(
            thrift.DataService,
            self.host,
            self.port,
            timeout=(self.request_timeout * 1000)
        )
        return self.client

    def ensure_auth(self):
        if self.inited:
            return

        if not self.username and not self.token:
            raise RuntimeError("未初始化")

        error, response = None, None
        for _ in range(self.request_attempt_count):
            try:
                self._create_client()  # 创建Thrift客户端
                if self.username:
                    # 用户名+密码登录
                    response = self.client.auth(self.username,
                                                self.password,
                                                self.compress,
                                                get_mac_address(),
                                                current_version, )
                else:
                    # Token登录
                    response = self.client.auth_by_token(self.token, self.audience)
                    break
            except socket_error as ex:
                error = ex
                time.sleep(0.5)
                if self.client:
                    self.client.close()
                    self.client = None
                continue
        else:
            if error and not response:
                raise error

        auth_message = zlib.decompress(response.msg).decode('utf-8')

        if not response.status:
            self._instances['_instance'] = None
            raise self.get_error(response)
        else:
            if self.not_auth:
                # log.debug("账号认证成功: %s" % auth_message)
                log.debug("账号认证成功")
                self.not_auth = False
        # 保存token
        self.token = json.loads(auth_message)['token']
        self.token_type = json.loads(auth_message)['token_type']
        self.inited = True

    def get_error(self, response):
        err = None
        if six.PY2:
            system = platform.system().lower()
            if system == "windows":
                err = Exception(response.error.encode("gbk"))
            else:
                err = Exception(response.error.encode("utf-8"))
        else:
            err = Exception(response.error)
        return err

    @classmethod
    def instance(cls):
        with cls._lock:
            _instance = cls._instances.get('_instance', None)
            if _instance is None:
                if cls._auth_params:
                    _instance = PQDataClient(**cls._auth_params)
                cls._instances['_instance'] = _instance
            return _instance

    def query(self, method, params):
        params["timeout"] = self.request_timeout
        params["request_id"] = next(self._request_id_generator)
        request = thrift.Query_Request()
        request.method_name = method
        request.params = msgpack.packb(params)
        buffer = six.BytesIO()
        result = None
        try:
            self.ensure_auth()
            response = self.client.query(request)
            if response.status:
                msg = response.msg
                if six.PY3 and isinstance(msg, str):
                    try:
                        msg = msg.encode("ascii")
                    except UnicodeError:
                        raise ResponseError("bad msg {!r}".format(msg))
                msg = zlib.decompress(msg)
                buffer.write(msg)
                pickle_encoding = None
                if six.PY3:
                    pickle_encoding = "latin1"
                result = pc.load(buffer, encoding=pickle_encoding)
            else:
                raise self.get_error(response)
        finally:
            buffer.close()

        if not isinstance(result, dict) or "request_id" not in result:
            return result

        if params["request_id"] != result["request_id"]:
            raise ResponseError("request_id {!r} != {!r}".format(
                params["request_id"], result["request_id"]
            ))
        return result["msg"]

    def __call__(self, method, *args, **kwargs):
        err, result = None, None
        max_retry = args[0] if len(args) > 0 else self.request_attempt_count  # 最多尝试次数
        result_on_except = args[1] if len(args) > 1 else None  # 如果尝试失败的返回默认结果
        retry_count = 0
        for _ in range(self.request_attempt_count):
            retry_count += 1
            try:
                result = self.query(method, kwargs)
                break
            except socket_error as ex:
                # 达到最大重试次数，返回异常时的默认结果
                if retry_count >= max_retry:
                    return result_on_except

                if not self._ping_server():
                    self._reset()
                err = ex
                time.sleep(0.6)
            except ResponseError as ex:
                err = ex

        if result is None and isinstance(err, Exception):
            raise err

        return self.convert_message(result)

    def __getattr__(self, method):
        # 创建一个闭包来捕获 name，并返回一个新的函数，该函数接受 *args 和 **kwargs
        def method_wrapper(*args, **kwargs):
            return self(method, *args, **kwargs)

            # 返回闭包
        return method_wrapper

    def _ping_server(self):
        if not self.client or not self.inited:
            return False
        for _ in range(self.request_attempt_count):
            try:
                msg = self.query("ping", {})
                return msg == "pong"
            except ResponseError:
                msg = None
                continue
            except Exception:
                return False

    @classmethod
    def convert_message(cls, msg):
        if isinstance(msg, dict):
            data_type = msg.get("data_type", None)
            data_value = msg.get("data_value", None)
            if data_type is not None and data_value is not None:
                params = data_value
                if data_type.startswith("pandas"):
                    data_index_type = params.pop("index_type", None)
                    if data_index_type == "Index":
                        params["index"] = pd.Index(params["index"])
                    elif data_index_type == "MultiIndex":
                        params["index"] = (
                            pd.MultiIndex.from_tuples(params["index"])
                            if len(params["index"]) > 0 else None
                        )
                    if data_type == "pandas_dataframe":
                        dtypes = params.pop("dtypes", None)
                        msg = pd.DataFrame(**params)
                        if dtypes:
                            msg = msg.astype(dtypes, copy=False)
                    elif data_type == "pandas_series":
                        msg = pd.Series(**params)
            else:
                msg = {key: cls.convert_message(val) for key, val in msg.items()}
        return msg

    def request(self, method, url, params=None, data=None, json=None, **kwargs):
        headers = {
            "username": self.username,
            "Authorization": f"{self.token_type} {self.token}"
        }
        return self.request_handler.request(method, url, params=params, data=data, json=json, headers=headers, **kwargs)




from six import with_metaclass
import abc
import datetime


class AbstractInstrument(with_metaclass(abc.ABCMeta)):
    """
    获取合约当前的实时基础信息和行情信息
    """

    @property
    @abc.abstractmethod
    def last_price(self) -> float:
        """
        最新价
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def high_limit(self) -> float:
        """
        涨停价
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def low_limit(self) -> float:
        """
        跌停价
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def paused(self) -> bool:
        """
        是否停止或者暂停了交易, 当停牌、未上市或者退市后返回 True
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def is_st(self) -> bool:
        """
        是否是 ST(包括ST, *ST)，是则返回 True，否则返回 False
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def day_open(self) -> float:
        """
        当天开盘价
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def display_name(self):
        """
        中文名称
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def name(self) -> str:
        """
        股票现在的名称, 可以用这个来判断股票当天是否是 ST, *ST, 是否快要退市
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def industry_name(self) -> str:
        """
        股票现在所属行业名称（申万L3）
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def industry_code(self) -> str:
        """
        股票现在所属行业代码（申万L3）
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def start_date(self) -> datetime.date:
        """
        上市日期， [datetime.date] 类型
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def end_date(self):
        """
        退市日期（股票是最后一个交易日，不同于摘牌日期）， [datetime.datetime] 类型, 如果没有退市则为2200-01-01
        :return:
        """
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def type(self):
        """
        证券类型
         'index'	#指数
         'stock'	#股票
         'fund'	    #基金
         'etf'		#ETF
        :return:
        """
        raise NotImplementedError
